/*************************************************************************/
/* DUCTAPE Version 1.0                                                   */
/* Copyright (C) 1998                                                    */
/* Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik */
/* Los Alamos National Laboratory, Advanced Computing Laboratory         */
/*************************************************************************/

#include "pdbType.h"
#include "pdbClass.h"
#include "pdbClassField.h"
#include "pdbRoutine.h"
#include "pdbNamespace.h"
#include "pdbTemplate.h"

ostream& operator<<(ostream& ostr, const pdbBase& b) {
  ostr << pdbItem::toName(b.virtuality())
       << " " << pdbItem::toName(b.access())
       << " cl#" << b.base()->id() << " ";
  return b.printLoc(ostr);
}
ostream& operator<<(ostream& ostr, const pdbFriendFunc& f) {
  ostr << "ro#" << f.friendFunc()->id() << " ";
  return f.printLoc(ostr);
}

ostream& operator<<(ostream& ostr, const pdbFriendClass& c) {
  ostr << "cl#" << c.friendClass()->id() << " ";
  return c.printLoc(ostr);
}

ostream& operator<<(ostream& ostr, const pdbFuncMember& f) {
  ostr << "ro#" << f.func()->id() << " ";
  return f.printLoc(ostr);
}

pdbClass::~pdbClass() {
  for (fieldvec::const_iterator dt=d.begin(); dt!=d.end(); ++dt)
     delete *dt;
  for (basevec::const_iterator bt=b.begin(); bt!=b.end(); ++bt)
     delete *bt;
  for (methodvec::const_iterator mt=f.begin(); mt!=f.end(); ++mt)
     delete *mt;
  for (friendfuncvec::const_iterator ft=fro.begin(); ft!=fro.end(); ++ft)
     delete *ft;
  for (friendclassvec::const_iterator ct=fcl.begin(); ct!=fcl.end(); ++ct)
     delete *ct;
}

pdbSimpleItem::dupl_t pdbClass::findDuplicate(pdbSimpleItem* r) {
  pdbClass* rhs = dynamic_cast<pdbClass*> (r);
  if ( fullName() == rhs->fullName() ) {
    int mysize = funcMembers().size() + dataMembers().size();
    int rhssize = rhs->funcMembers().size() + rhs->dataMembers().size();
    if ( mysize < rhssize )
      return OLDDUPL;
    else
      return NEWDUPL;
  } else {
      return NODUPL;
  }
}

void pdbClass::process(PDB *p) {
  // ??? name is fullName, because of bug in frontend
  // ??? to fix this bug, try to correct name
  size_t pos;
  if ( !isTemplate() ) {
    if ( (pos = name().rfind("::")) != STR_NPOS ) {
      name(name().substr(pos+2, STR_NPOS));
    }
    pdbItem::process(p);
  } else {
    fullName(name());
  }
  // the same for type members
  const pdbType* ty;
  for( fieldvec::iterator cit = d.begin(); cit!=d.end(); ++cit) {
    if ( (*cit)->kind() == M_TYPE ) {
      ty = (*cit)->type();
      if ( ty->kind()==TY_ENUM || ty->kind()==TY_CLASS ||
           (ty->kind()==TY_TREF && ty->qualifiers().size()==0) ) {
        if ( (pos = (*cit)->name().rfind("::")) != STR_NPOS ) {
          (*cit)->name((*cit)->name().substr(pos+2, STR_NPOS));
        }
      }
    }
  }

  // if class has no base classes, add it to the toplevel classes
  if ( b.size() == 0 ) {
    p->classTree()->addDerivedClass(this);
  }

  // determine if class is Abstract Base Class
  const pdbClass::methodvec f = funcMembers();
  for (pdbClass::methodvec::const_iterator ft=f.begin(); ft!=f.end(); ++ft) {
    if ( (*ft)->func()->virtuality() == VI_PURE ) {
      abc = true;
      break;
    }
  }
}

ostream& pdbClass::print(ostream& ostr) const {
  pdbItem::print(ostr);
  ostr << "ckind " << toName(kind()) << "\n";
  pdbTemplateItem::print(ostr);
  const pdbClass::basevec b = baseClasses();
  for (pdbClass::basevec::const_iterator bt=b.begin(); bt!=b.end(); ++bt)
     ostr << "cbase " << (**bt) << "\n";
  const pdbClass::friendclassvec fc = friendClasses();
  for (pdbClass::friendclassvec::const_iterator fct=fc.begin();
       fct!=fc.end(); ++fct)
     ostr << "cfrclass " << (**fct) << "\n";
  const pdbClass::friendfuncvec fr = friendRoutines();
  for (pdbClass::friendfuncvec::const_iterator frt=fr.begin();
       frt!=fr.end(); ++frt)
     ostr << "cfrfunc " << (**frt) << "\n";
  const pdbClass::methodvec f = funcMembers();
  for (pdbClass::methodvec::const_iterator ft=f.begin(); ft!=f.end(); ++ft)
     ostr << "cfunc " << (**ft) << "\n";
  const pdbClass::fieldvec d = dataMembers();
  for (pdbClass::fieldvec::const_iterator dt=d.begin(); dt!=d.end(); ++dt)
     ostr << **dt;
  pdbFatItem::print(ostr);
  return ostr << endl;
}

void pdbClass::adjustPtrs(PDB* p) {
  pdbClassField* fd;
  const pdbFile* fi;
  const pdbTemplate *tm;

  pdbTemplateItem::adjustPtrs(p);
  for( basevec::iterator bit = b.begin(); bit!=b.end(); ++bit) {
    if ( (*bit)->base()->newId() > UNIQUE )
      (*bit)->base(p->getClassMap()[(*bit)->base()->newId()]);
    if ( (fi=(*bit)->file()) && (fi->newId() > UNIQUE) )
      (*bit)->file(p->getFileMap()[fi->newId()]);
  }
  for( friendclassvec::iterator fcit = fcl.begin(); fcit!=fcl.end(); ++fcit) {
    if ( (*fcit)->friendClass()->newId() > UNIQUE )
      (*fcit)->friendClass(p->getClassMap()[(*fcit)->friendClass()->newId()]);
    if ( (fi=(*fcit)->file()) && (fi->newId() > UNIQUE) )
      (*fcit)->file(p->getFileMap()[fi->newId()]);
  }
  for( friendfuncvec::iterator frit = fro.begin(); frit!=fro.end(); ++frit) {
    if ( (*frit)->friendFunc()->newId() > UNIQUE )
      (*frit)->friendFunc(p->getRoutineMap()[(*frit)->friendFunc()->newId()]);
    if ( (fi=(*frit)->file()) && (fi->newId() > UNIQUE) )
      (*frit)->file(p->getFileMap()[fi->newId()]);
  }
  for( classvec::iterator sit = s.begin(); sit!=s.end(); ++sit) {
    if ( (*sit)->newId() > UNIQUE )
      (*sit) = p->getClassMap()[(*sit)->newId()];
  }
  for( methodvec::iterator fit = f.begin(); fit!=f.end(); ++fit) {
    if ( (*fit)->func()->newId() > UNIQUE )
      (*fit)->func(p->getRoutineMap()[(*fit)->func()->newId()]);
    if ( (fi=(*fit)->file()) && (fi->newId() > UNIQUE) )
      (*fit)->file(p->getFileMap()[fi->newId()]);
  }
  for( fieldvec::iterator cit = d.begin(); cit!=d.end(); ++cit) {
    fd = (*cit);
    // if ( fd->parentNSpace() && fd->parentNSpace()->newId() > UNIQUE )
    //   fd->parentNSpace(p->getNamespaceMap()[fd->parentNSpace()->newId()]);
    if ( fd->parentClass()->newId() > UNIQUE )
      fd->parentClass(p->getClassMap()[fd->parentClass()->newId()]);
    if ( fd->type()->newId() > UNIQUE )
      fd->type(p->getTypeMap()[fd->type()->newId()]);
    if ( (fi=fd->location().file()) && (fi->newId() > UNIQUE) )
      fd->location().file(p->getFileMap()[fi->newId()]);
    if ( (tm=fd->isTemplate()) && (tm != TE_UNKNOWN) && (tm->newId() > UNIQUE) )
      fd->isTemplate(p->getTemplateMap()[tm->newId()]);
  }
}
