/*************************************************************************/
/* DUCTAPE Version 1.0                                                   */
/* Copyright (C) 1998                                                    */
/* Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik */
/* Los Alamos National Laboratory, Advanced Computing Laboratory         */
/*************************************************************************/

#ifdef _OLD_HEADER_
# include <fstream.h>
# include <algo.h>
#else
# include <fstream>
  using std::ifstream;
  using std::ios;
# include <algorithm>
  using std::remove_if;
  using std::find_if;
#endif
#include <stdlib.h>
#include "pdbAll.h"

static const char *fname = 0;
static int line = -1;
static bool ok = true;

void ERROR(const char *msg, const char *val) {
  cerr << "Error: " << fname << ":" << line << ": " << msg << val << endl;
  ok = false;
}

void ERROR(const char *msg, const char *item, int val) {
  cerr << "Error: " << fname << ":" << line << ": "
       << msg << " " << item << " "
       << item[0] << item[1] << "#" << val << endl;
  ok = false;
}

#include "pdbTraits.h"

template<class tag>
typename PDBTraits<tag>::item_t *PDB::findItem(tag t, int id) {
  typename PDBTraits<tag>::map_t& map = PDBTraits<tag>::getMap(this);
  typename PDBTraits<tag>::map_t::iterator it = map.find(id);
  if ( it != map.end() )
    return (*it).second;
  else {
    PDBTraits<tag>::item_t* item = new PDBTraits<tag>::item_t(id);
    map[id] = item;
    PDBTraits<tag>::getItems(this).push_back(item);
    return item;
  }
}

template<class tag>
typename PDBTraits<tag>::item_t* PDB::findItem(tag t, const string& name, int id) {
  typename PDBTraits<tag>::map_t& map = PDBTraits<tag>::getMap(this);
  typename PDBTraits<tag>::map_t::iterator it = map.find(id);
  if ( it != map.end() ) {
    PDBTraits<tag>::item_t *i = (*it).second;
    if ( i->name()[0] ) ERROR("Duplicate", i->desc(), id);
    i->name(name);
    return i;
  } else {
    PDBTraits<tag>::item_t* item = new PDBTraits<tag>::item_t(name, id);
    map[id] = item;
    PDBTraits<tag>::getItems(this).push_back(item);
    return item;
  }
}

template<class tag>
void PDB::finalCheck(tag t) {
  typename PDBTraits<tag>::map_t& map = PDBTraits<tag>::getMap(this);
  typename PDBTraits<tag>::vec_t& vec = PDBTraits<tag>::getVec(this);
  for (typename PDBTraits<tag>::map_t::const_iterator mit = map.begin();
       mit != map.end(); ++mit) {
    if (! (*mit).second->name()[0] )
      ERROR("Undefined", (*mit).second->desc(), (*mit).second->id());
    (*mit).second->process(this);
    vec.push_back((*mit).second);
  }
}

template<class tag>
void PDB::markDuplicates(tag t, PDB& other) {
  typename PDBTraits<tag>::vec_t& vecOld = PDBTraits<tag>::getVec(this);
  typename PDBTraits<tag>::vec_t& vecNew = PDBTraits<tag>::getVec(&other);
  typename PDBTraits<tag>::map_t& mapOld = PDBTraits<tag>::getMap(this);
  typename PDBTraits<tag>::map_t& mapNew = PDBTraits<tag>::getMap(&other);
  int maxId = 0;
  bool oldDupls = false;

  // look for duplicate item in "other" PDB
  // use "findDuplicate" to check for duplicates
  for (typename PDBTraits<tag>::vec_t::const_iterator oit = vecOld.begin();
       oit != vecOld.end(); ++oit) {
    for (typename PDBTraits<tag>::vec_t::const_iterator nit = vecNew.begin();
         nit != vecNew.end(); ++nit) {
      // only compare Items if not marked as duplicate yet
      if ( (*nit)->newId() == pdbSimpleItem::UNIQUE ) {
        pdbSimpleItem::dupl_t res = (*oit)->findDuplicate(*nit);
        if ( res == pdbSimpleItem::OLDDUPL ) {
          // duplicate in "old" PDB 
          (*oit)->newId((*nit)->id());  // "old" id, needs to be fixed later
          oldDupls = true;
          break;
        } else if ( res == pdbSimpleItem::NEWDUPL ) {
          // duplicate in "other" PDB
          (*nit)->newId((*oit)->id());
          break;
        }
      }
    }
    // determine highest used ID
    if ( (*oit)->id() > maxId ) maxId = (*oit)->id();
  }

  // assign new items (non duplicates) from "other" PDB new ID
  // and insert them in "old" PDB (map AND vecs!)
  for (typename PDBTraits<tag>::vec_t::const_iterator nit = vecNew.begin();
       nit != vecNew.end(); ++nit) {
    if ( (*nit)->newId() == pdbSimpleItem::UNIQUE ) {
      ++maxId;
      mapOld[maxId] = (*nit);
      vecOld.push_back(*nit);
      PDBTraits<tag>::getItems(this).push_back(*nit);
      (*nit)->id(maxId);

      // mark them, so they are not deleted if "other" PDB is destructed
      (*nit)->newId(pdbSimpleItem::NOTDEL);
    }
  }

  // if duplicates in "old" PDB, fix the references to the new id's
  if ( oldDupls ) {
    for (typename PDBTraits<tag>::vec_t::const_iterator oit = vecOld.begin();
         oit != vecOld.end(); ++oit) {
      if ( (*oit)->newId() > pdbSimpleItem::UNIQUE ) {
         (*oit)->newId((mapNew[(*oit)->newId()])->id());
      }
    }
  }
}

static bool isDuplVec(const pdbSimpleItem* s) {
  return s->newId() > pdbSimpleItem::UNIQUE;
}

static bool isRoDuplMap(map<int, pdbRoutine*>::value_type& r) {
  return (r.second)->newId() > pdbSimpleItem::UNIQUE;
}

static bool isClDuplMap(map<int, pdbClass*>::value_type& r) {
  return (r.second)->newId() > pdbSimpleItem::UNIQUE;
}


bool PDB::merge(char *fname) {
  {
    PDB other(fname);
    if (!other) return false;

    markDuplicates(typeTag(), other);
    markDuplicates(fileTag(), other);
    markDuplicates(classTag(), other);
    markDuplicates(routineTag(), other);
    markDuplicates(templateTag(), other);
    markDuplicates(macroTag(), other);
    markDuplicates(namespaceTag(), other);

    // update/merge callTree / fileTree / classTree
    for( pdbRoutine::callvec::iterator
         orit = other.topRoutine->callees().begin();
         orit!=other.topRoutine->callees().end(); ++orit) {
      const pdbRoutine *r = (*orit)->call();
      if ( r->newId() <= pdbSimpleItem::UNIQUE )
        topRoutine->addCallee(r, pdbItem::VI_NO, 0, 0, 0);
    }
    for( pdbFile::incvec::iterator ofit = other.topFile->includes().begin();
         ofit!=other.topFile->includes().end(); ++ofit) {
      if ( (*ofit)->newId() <= pdbSimpleItem::UNIQUE )
        topFile->addInclude(*ofit);
    }
    for( pdbClass::classvec::iterator
         ocit = other.topClass->derivedClasses().begin();
         ocit!=other.topClass->derivedClasses().end(); ++ocit) {
      if ( (*ocit)->newId() <= pdbSimpleItem::UNIQUE )
        topClass->addDerivedClass(*ocit);
    }

    // adjust pointer in items which ID didn't change
    for (itemvec::iterator it=itemVec.begin(); it!=itemVec.end(); ++it) {
      if ( (*it)->newId() <= pdbSimpleItem::UNIQUE ) (*it)->adjustPtrs(this);
    }

    // adjust pointers in callTree / fileTree / classTree too
    for( pdbRoutine::callvec::iterator rit = topRoutine->callees().begin();
         rit!=topRoutine->callees().end(); ++rit) {
      const pdbRoutine *r = (*rit)->call();
      if ( r->newId() > pdbSimpleItem::UNIQUE )
        (*rit)->call(getRoutineMap()[r->newId()]);
    }
    for( pdbFile::incvec::iterator fit = topFile->includes().begin();
         fit!=topFile->includes().end(); ++fit) {
      if ( (*fit)->newId() > pdbSimpleItem::UNIQUE )
        *fit = getFileMap()[(*fit)->newId()];
    }
    for( pdbClass::classvec::iterator cit = topClass->derivedClasses().begin();
         cit!=topClass->derivedClasses().end(); ++cit) {
      if ( (*cit)->newId() > pdbSimpleItem::UNIQUE )
        *cit = getClassMap()[(*cit)->newId()];
    }

    // remove duplicates (in "this" PDB)
    // only need to do it for routines and classes
    itemVec.erase(remove_if(itemVec.begin(), itemVec.end(), isDuplVec),
                  itemVec.end());
    routineVec.erase(remove_if(routineVec.begin(), routineVec.end(), isDuplVec),
                     routineVec.end());
    classVec.erase(remove_if(classVec.begin(), classVec.end(), isDuplVec),
                   classVec.end());

    routinemap::iterator rrit = routineMap.begin(), rnit;
    while ( (rrit = find_if(rrit, routineMap.end(), isRoDuplMap))
            != routineMap.end() ) {
      rnit = rrit;
      ++rrit;
      routineMap.erase(rnit);
    }
    classmap::iterator ccit = classMap.begin(), cnit;
    while ( (ccit = find_if(ccit, classMap.end(), isClDuplMap))
            != classMap.end() ) {
      cnit = ccit;
      ++ccit;
      classMap.erase(cnit);
    }
  }

  // reset newId
  for (itemvec::iterator it=itemVec.begin(); it!=itemVec.end(); ++it) {
    (*it)->newId(pdbSimpleItem::UNIQUE);
  }

  return true;
}

PDB::attr_t PDB::toAttr(const char *v) {
  attrmap::iterator it = attr.find(v);
  if ( it == attr.end() ) {
    ERROR("Unknown attribute ", v);
    return NOATTR;
  }
  return (*it).second;
}

pdbType *PDB::getTypeOrClass(const char *value) {
  if ( value[0] == 't' && value[1] == 'y' ) {
    return findItem(typeTag(), atoi(value+3));
  } else if ( value[0] == 'c' && value[1] == 'l' ) {
    pdbType *t = findItem(typeTag(), -atoi(value+3));
    t->name(value);
    t->kind(pdbItem::TY_CLASS);
    t->isClass(findItem(classTag(), atoi(value+3)));
    return t;
  } else {
    ERROR("Bad typeID ", value);
    return 0;
  }
}

bool PDB::getLocation(const char *value, char **ptr,
                      pdbFile*& file, int& line, int& col) {

  if ( value[0] == 's' && value[1] == 'o' ) {
    int f = strtol(value+3, ptr, 10);
    line = strtol((*ptr)+1, ptr, 10);
    col  = strtol((*ptr)+1, ptr, 10);
    file = findItem(fileTag(), f);
  } else if (  value[0]=='N' && value[1]=='U' && value[2]=='L' ) {
    *ptr += 9;
    file = 0;
    line = col = 0;
  } else {
    ERROR("Bad fileID ", value);
    return false;
  }
  return true;
}

const int INBUF_SIZE=2048;

bool PDB::read(char *filename) {
  ifstream istr(filename);
  static char inbuf[INBUF_SIZE];
  bool expectNew = true;
  char *value;
  char currItem = '\0';
  long num;
  int l, c;
  pdbFile *f = 0;
  pdbItem *iptr = 0;
  pdbClass *cptr = 0;
  pdbClassField *cmptr = 0;
  pdbRoutine *rptr = 0;
  pdbTemplate *tptr = 0;
  pdbTemplateItem *tiptr = 0;
  pdbMacro *mptr = 0;
  pdbType *yptr = 0;
  pdbNamespace *nptr = 0;
  pdbFile *fptr = 0;
  pdbFatItem *fiptr = 0;

  fname = filename;
  line = 1;
  ok = true;
  inbuf[INBUF_SIZE-1] = '\n';

  if (!istr) {
    if ( fname )
      cerr << "Error: Cannot open '" << fname << "'" << endl;
    else
      cerr << "Error: Cannot open ''" << endl;
    return false;
  }

  // check magic number
  if ( !istr.getline(inbuf, INBUF_SIZE) || (strcmp(inbuf, "<PDB 1.0>")!=0) ) {
    cerr << "Error: Wrong magic number" << endl;
    return false;
  } else {
    // skip empty line
    istr.getline(inbuf, INBUF_SIZE);
  }
  
  while ( istr.getline(inbuf, INBUF_SIZE) || istr.gcount() ) {
    if ( inbuf[0] == '\0' ) {
      // empty line -> end of item
      // complete / check item
      expectNew = true;
      iptr = 0;
      cptr = 0;
      cmptr = 0;
      rptr = 0;
      tptr = 0;
      mptr = 0;
      nptr = 0;
      fptr = 0;
      yptr = 0;
      tiptr = 0;
      fiptr = 0;

    } else if ( expectNew ) {
      // New Item
      char fst = currItem = inbuf[0];
      char snd = inbuf[1];
      num = strtol(inbuf+3, &value, 10);
      string name( (*value == '\0' || *(value+1) == '\0') ? "-" : ++value);

      // check for very long lines
      while ( inbuf[INBUF_SIZE-1] != '\n' ) {
        inbuf[INBUF_SIZE-1] = '\n';
        if (istr.fail()) istr.clear(istr.rdstate() & ~ios::failbit);     
        istr.getline(inbuf, INBUF_SIZE);
        name += inbuf;
      }

      if ( fst=='r' && snd=='o' ) {
        iptr = fiptr = tiptr = rptr = findItem(routineTag(), name, num);
      } else if ( fst=='s' && snd=='o' ) {
        fptr = findItem(fileTag(), name, num);
      } else if ( fst=='c' && snd=='l' ) {
        iptr = fiptr = tiptr = cptr = findItem(classTag(), name, num);
      } else if ( fst=='t' && snd=='y' ) {
        iptr = yptr = findItem(typeTag(), name, num);
        currItem = 'y';
      } else if ( fst=='t' && snd=='e' ) {
        iptr = fiptr = tptr = findItem(templateTag(), name, num);
      } else if ( fst=='m' && snd=='a' ) {
        iptr = mptr = findItem(macroTag(), name, num);
      } else if ( fst=='n' && snd=='a' ) {
        iptr = fiptr = nptr = findItem(namespaceTag(), name, num);
      } else {
        ERROR("Unknown item ", inbuf);
        currItem = '\0';
      }
      expectNew = false;

    } else {
      // Attributes
      char *ptr = strchr(inbuf, ' ');
      *ptr = '\0';
      value = ptr+1;
      attr_t a = toAttr(inbuf);
      if ( currItem != inbuf[0] ) {
        ERROR("Non-matching attribute ", inbuf);
      } else {
        switch (a) {
        // ==== Common Attributes =======================================
        case ACS:      // ---- Access
          if (iptr) iptr->access(pdbItem::toAccess(value));
          break;
        case CLASS:    // ---- Parent Class
	  if ( value[0] == 'c' && value[1] == 'l' ) {
            if (iptr) iptr->parentClass(findItem(classTag(), atoi(value+3)));
	  } else ERROR("Bad classID ", value);
          break;
        case LOC:      // ---- Location: fileId Line Column
          if ( getLocation(value, &ptr, f, l, c) && iptr )
            iptr->location(pdbLoc(f, l, c));
          break;
        case POS: {    // ---- Position: loc1 loc2 loc3 loc4
                       // ----           locN := file line col
            if ( getLocation(value, &ptr, f, l, c) )
              fiptr->headBegin(pdbLoc(f, l, c));
            if ( getLocation(ptr+1, &ptr, f, l, c) )
              fiptr->headEnd  (pdbLoc(f, l, c));
            if ( getLocation(ptr+1, &ptr, f, l, c) )
              fiptr->bodyBegin(pdbLoc(f, l, c));
            if ( getLocation(ptr+1, &ptr, f, l, c) )
              fiptr->bodyEnd  (pdbLoc(f, l, c));
          }
          break;
        case NSPACE:   // ---- Parent Namespace
	  if ( value[0] == 'n' && value[1] == 'a' ) {
            if (iptr)
              iptr->parentNSpace(findItem(namespaceTag(), atoi(value+3)));
	  } else ERROR("Bad namespaceID ", value);
          break;

        // ==== Common Template Item Attributes =========================
        case TEMPL:    // ---- Template
	  if ( value[0] == 't' && value[1] == 'e' ) {
            if (tiptr)
              tiptr->isTemplate(findItem(templateTag(), atoi(value+3)));
	  } else if ( value[0] == 'T' ) {
            if (tiptr)
              tiptr->isTemplate(pdbTemplateItem::TE_UNKNOWN);
          } else ERROR("Bad templateID ", value);
          break;
        case SPECL:    // ---- Is Specialized?
          if (tiptr) tiptr->isSpecialized(value[0] == 'T');
          break;

        // ==== Class Attributes ========================================
        case CBASE: {  // ---- Direct Base Class
          ptr = strtok(value, " ");
          pdbItem::virt_t v = pdbItem::toVirt(ptr);
          ptr = strtok(NULL, " ");
          pdbItem::access_t a = pdbItem::toAccess(ptr);
          ptr = strtok(NULL, " ");
	  if ( ptr[0] == 'c' && ptr[1] == 'l' ) {
            if (cptr) {
              int b = strtol(ptr+3, &ptr, 10);
              if ( getLocation(ptr+1, &ptr, f, l, c) ) {
                pdbClass *cb = findItem(classTag(), b);
                cptr->addBaseClass(v, a, cb, f, l, c);
                cb->addDerivedClass(cptr);
              }
            }
	  } else ERROR("Bad classID ", ptr);
          }
          break;
        case CFUNC:    // ---- Member Function
	  if ( value[0] == 'r' && value[1] == 'o' ) {
            if (cptr) {
              int r = strtol(value+3, &ptr, 10);
              if ( getLocation(ptr+1, &ptr, f, l, c) ) {
                cptr->addFuncMember(findItem(routineTag(), r), f, l, c);
              }
            }
	  } else ERROR("Bad routineID ", value);
          break;
        case CKIND:    // ---- Class Kind
          if (cptr) cptr->kind(pdbItem::toClass(value));
          break;
        case CMEM: {   // ---- Data Member
          string name(value);

          // check for very long lines
          while ( inbuf[INBUF_SIZE-1] != '\n' ) {
            inbuf[INBUF_SIZE-1] = '\n';
            if (istr.fail()) istr.clear(istr.rdstate() & ~ios::failbit);     
            istr.getline(inbuf, INBUF_SIZE);
            name += inbuf;
          }

          iptr = tiptr = cmptr = new pdbClassField(name);
          if (cptr) {
            cptr->addDataMember(cmptr);
            cmptr->parentClass(cptr);
          }
          }
          break;
       case CFRCLASS: // ---- Friend Class
	  if ( value[0] == 'c' && value[1] == 'l' ) {
            if (cptr) {
              int b = strtol(value+3, &ptr, 10);
              if ( getLocation(ptr+1, &ptr, f, l, c) ) {
                cptr->addFriendClass(findItem(classTag(), b), f, l, c);
              }
            }
	  } else ERROR("Bad classID ", value);
          break;
        case CFRFUNC:  // ---- Friend Function
	  if ( value[0] == 'r' && value[1] == 'o' ) {
            if (cptr) {
              int b = strtol(value+3, &ptr, 10);
              if ( getLocation(ptr+1, &ptr, f, l, c) ) {
                cptr->addFriendRoutine(findItem(routineTag(), b), f, l, c);
              }
            }
	  } else ERROR("Bad classID ", value);
          break;

        // ==== Class Field Attributes ==================================
        case CMKIND:   // ---- Class Field Kind
          if (cmptr) cmptr->kind(pdbItem::toMem(value));
          break;
        case CMISBIT:  // ---- Is Bitfield?
          if (cmptr) cmptr->isBitField(value[0] == 'T');
          break;
        case CMTYPE:   // ---- type
          if (cmptr) cmptr->type(getTypeOrClass(value));
          break;
        case CMCONST:  // ---- Is Static Constant?
          if (cmptr) cmptr->isStaticConst(value[0] == 'T');
          break;
        case CMMUT:    // ---- Is Mutable?
          if (cmptr) cmptr->isMutable(value[0] == 'T');
          break;

        // ==== File Attributes =========================================
        case SINC:     // ---- File Include
	  if ( value[0] == 's' && value[1] == 'o' ) {
            if (fptr) {
              fptr->addInclude(findItem(fileTag(), atoi(value+3)));
            }
	  } else ERROR("Bad fileID ", value);
          break;

        // ==== Macro Attributes ========================================
        case MKIND:    // ---- Macro Kind
          if (mptr) mptr->kind(pdbItem::toMacro(value));
          break;
        case MTEXT:    // ---- Macro Text
          if (mptr) {
            string txt(value);
            while ( inbuf[INBUF_SIZE-1] != '\n' ) {
              inbuf[INBUF_SIZE-1] = '\n';
              if (istr.fail()) istr.clear(istr.rdstate() & ~ios::failbit);     
              istr.getline(inbuf, INBUF_SIZE);
              txt += inbuf;
            }
            mptr->text(txt);
          }
          break;

        // ==== Routine Attributes ======================================
        case RCALL:    // ---- Calls Routine
	  if ( value[0] == 'r' && value[1] == 'o' ) {
            int r = strtol(value+3, &ptr, 10);
            ptr = strtok(ptr+1, " ");
            pdbItem::virt_t v = pdbItem::toVirt(ptr);
            ptr = strtok(NULL, " ");
            if ( getLocation(ptr, &ptr, f, l, c) && rptr ) {
              rptr->addCallee(findItem(routineTag(), r), v, f, l, c);
            }
	  } else ERROR("Bad routineID ", value);
          break;
        case RCRVO:    // ---- Covariant Return Virtual Override?
          if (rptr) rptr->covariantReturnVirtualOverride(value[0] == 'T');
          break;
        case RINLINE:  // ---- Is Inline?
          if (rptr) rptr->isInline(value[0] == 'T');
          break;
        case RKIND:    // ---- Special Kind
          if (rptr) rptr->specialKind(pdbItem::toFunc(value));
          break;
        case RSIG:     // ---- Signature
	  if ( value[0] == 't' && value[1] == 'y' ) {
            if (rptr) rptr->signature(findItem(typeTag(), atoi(value+3)));
	  } else ERROR("Bad typeID ", value);
          break;
        case RLINK:    // ---- Linkage
          if (rptr) rptr->linkage(pdbItem::toLink(value));
          break;
        case RSTAT:  // ---- Is Static?
          if (rptr) rptr->isStatic(value[0] == 'T');
          break;
        case RSTORE:   // ---- Storage Class
          if (rptr) rptr->storageClass(pdbItem::toStore(value));
          break;
        case RVIRT:    // ---- Virtuality
          if (rptr) rptr->virtuality(pdbItem::toVirt(value));
          break;
        case RCGEN:    // ---- Is Compiler Generated?
          if (rptr) rptr->isCompilerGenerated(value[0] == 'T');
          break;
        case REXPL:    // ---- Is Explicit Constructor?
          if (rptr) rptr->isExplicitCtor(value[0] == 'T');
          break;

        // ==== Template Attributes =====================================
        case TKIND:    // ---- Template Kind
          if (tptr) tptr->kind(pdbItem::toTempl(value));
          break;
        case TTEXT:    // ---- Template Text
          if (tptr) {
            string txt(value);
            while ( inbuf[INBUF_SIZE-1] != '\n' ) {
              inbuf[INBUF_SIZE-1] = '\n';
              if (istr.fail()) istr.clear(istr.rdstate() & ~ios::failbit);     
              istr.getline(inbuf, INBUF_SIZE);
              txt += inbuf;
            }
            tptr->text(txt);
          }
          break;

        // ==== Namespace Attributes ====================================
        case NALIAS:   // ---- Alias Name
	  if ( value[0] == 'n' && value[1] == 'a' ) {
            if (nptr) nptr->isAlias(findItem(namespaceTag(), atoi(value+3)));
	  } else ERROR("Bad namespaceID ", value);
          break;
        case NMEM:     // ---- Namespace Member
          if ( value[0]=='r' && value[1]=='o' ) {
            if (nptr) nptr->addMember(findItem(routineTag(), atoi(value+3)));
          } else if ( value[0]=='c' && value[1]=='l' ) {
            if (nptr) nptr->addMember(findItem(classTag(), atoi(value+3)));
          } else if ( value[0]=='t' && value[1]=='y' ) {
            if (nptr) nptr->addMember(findItem(typeTag(), atoi(value+3)));
          } else if ( value[0]=='t' && value[1]=='e' ) {
            if (nptr) nptr->addMember(findItem(templateTag(), atoi(value+3)));
          } else if ( value[0]=='n' && value[1]=='a' ) {
            if (nptr) nptr->addMember(findItem(namespaceTag(), atoi(value+3)));
          } else {
            ERROR("Unknown item ", value);
          }
          break;

        // ==== Type Attributes =========================================
        case YELEM:    // ---- Element Type (if Array)
          if (yptr) yptr->elementType(getTypeOrClass(value));
          break;
        case YNELEM:   // ---- Number of Elements (if Array)
          if (yptr) yptr->numElements(atoi(value));
          break;
        case YFKIND:   // ---- Float Kind (if Float)
          if (yptr) yptr->floatType(pdbItem::toFloat(value));
          break;
        case YIKIND:   // ---- Integer Kind (if Integer)
          if (yptr) yptr->integerType(pdbItem::toInt(value));
          break;
        case YKIND:    // ---- Kind of Type
          if (yptr) yptr->kind(pdbItem::toType(value));
          break;
        case YQUAL:    // ---- Type Qualifiers
          ptr = strtok(value, " ");
          while ( ptr != NULL ) {
            if (yptr) yptr->addQualifier(pdbItem::toQual(ptr));
            ptr = strtok(NULL, " ");
          }
          break;
        case YPTR:     // ---- Type pointed to (if Pointer)
          if (yptr) yptr->pointedToType(getTypeOrClass(value));
          break;
        case YREF:     // ---- Type referenced (if Reference)
          if (yptr) yptr->referencedType(getTypeOrClass(value));
          break;
        case YSIGNED:  // ---- is explicitly Signed?
          if (yptr) yptr->isSigned(value[0] == 'T');
          break;
        case YTREF:    // ---- Base Type (if Typedef or Qualified Type)
          if (yptr) yptr->baseType(getTypeOrClass(value));
          break;
        case YRETT:    // ---- Return Type
          if (yptr) yptr->returnType(getTypeOrClass(value));
          break;
        case YARGT:    // ---- Argument Type
          if (yptr) {
            ptr = strtok(value, " ");
            ptr = strtok(NULL, " ");
            yptr->addArgumentType(getTypeOrClass(value), (ptr && ptr[0]=='T'));
          }
          break;
        case YELLIP:   // ---- Has Ellipsis?
          if (yptr) yptr->hasEllipsis(value[0] == 'T');
          break;
        case YMPCLASS: // ---- Member Pointer Parent Class
	  if ( value[0] == 'c' && value[1] == 'l' ) {
            if (yptr)
              yptr->memberPointerClass(findItem(classTag(), atoi(value+3)));
	  } else ERROR("Bad classID ", value);
          break;
        case YMPTYPE:  // ---- Member Pointer Type
          if (yptr) yptr->memberPointerType(getTypeOrClass(value));
          break;
        case YEXCEP:   // ---- Exception Specification
	  if ( value[0] == 'c' && value[1] == 'l' ) {
            if (yptr)
              yptr->addExceptionSpec(findItem(classTag(), atoi(value+3)));
          } else if ( strcmp(value, "NULL") == 0 ) {
              yptr->addExceptionSpec(pdbType::NULLEXCEP);
	  } else ERROR("Bad classID ", value);
          break;
        case YENUM:    // ---- Enumeration Value
          ptr = strtok(value, " ");
          ptr = strtok(NULL, " ");
          *(ptr-1) = '\0';
          yptr->addEnumeration(pdbEnum(value, strtol(ptr, 0, 0)));
          break;

        // ==== Invalid Attributes= =====================================
        case NOATTR:
          break;

        default:
          ERROR("*** INTERNAL PROGRAM ERROR *** at ", value);
          exit(1);
        }
      }
    }
    if ( inbuf[INBUF_SIZE-1] != '\n' ) {
      ERROR("*** LINE TOO LONG ***", "");
      cerr << inbuf << endl;
      exit(1);
    }
    line++;
    inbuf[INBUF_SIZE-1] = '\n';
  }

  // Are items missing?
  // - build full names and other item specific finalization
  finalCheck(typeTag());
  finalCheck(fileTag());
  finalCheck(classTag());
  finalCheck(templateTag());
  finalCheck(macroTag());
  finalCheck(namespaceTag());
  if (!ok) return false;
  finalCheck(routineTag());
  return ok;
}
