/*************************************************************************/
/* DUCTAPE Version 1.0                                                   */
/* Copyright (C) 1998                                                    */
/* Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik */
/* Los Alamos National Laboratory, Advanced Computing Laboratory         */
/*************************************************************************/

#include <ctype.h>
#include "pdbType.h"
#include "pdbRoutine.h"
#include "pdbClass.h"

ostream& operator<<(ostream& ostr, const pdbCallee& c) {
  ostr << "ro#" << c.call()->id() << " "
       << pdbItem::toName(c.virtuality()) << " ";
  return c.printLoc(ostr);
}

pdbRoutine::~pdbRoutine() {
  for (callvec::const_iterator ct=c.begin(); ct!=c.end(); ++ct) delete *ct;
}

void pdbRoutine::incrCalled() const {
  pdbRoutine *self = const_cast<pdbRoutine*>(this);
  self->n++;
}

void pdbRoutine::process(PDB *p) {
  // if routine wasn't called, add it to the toplevel functions
  if ( numCalled() == 0 ) {
    p->callTree()->addCallee(this, pdbItem::VI_NO, 0, 0, 0);
  }

  // routines without location or the same location as the parent class
  // are computer generated ???
  if ( !location().file() ||
       (parentClass() && parentClass()->location() == location()) ) {
    if ( !isCGen ) { 
      //cerr << "**** FOUND ANOTHER CGEN" << endl;
      //cerr << "ro#" << id() << " " << name() << " file=" << (location().file()==0) << endl;
      isCGen = true;
    }
  }

  // build up full name:
  // do it like for pdbItem, but add signature
  pdbItem::process(p);


  // Implementation A:
  // build it out of return type and argument types
  const pdbType *ty = signature()->returnType();
  string other = ty->isClass() ? ty->isClass()->fullName() : ty->fullName();
  string leftover = "";
  if ( isalnum(other[other.size()-1]) || other[other.size()-1]=='>' )
    other += " ";
  if ( signature()->returnType()->kind() == TY_PTR &&
       signature()->returnType()->pointedToType()->kind() == TY_FUNC ) {
    // return type is function pointer
    const char *ret = other.c_str();
    const char *ptr = strstr(ret, "(*");
    int idx = (ptr - ret) + 2;
    int len = other.size() - 1;
    leftover = other.substr(idx, len-idx+1);
    other = other.substr(0,idx) + fullName() + "(";
  } else {
    other += fullName() + "(";
  }
  for (pdbType::argvec::const_iterator ait=signature()->argumentTypes().begin();
       ait!=signature()->argumentTypes().end(); ++ait) {
    if ( ait!=signature()->argumentTypes().begin() ) other += ", ";
    other += ait->type()->isClass() ?
             ait->type()->isClass()->fullName() : ait->type()->fullName();
  }
  if ( signature()->hasEllipsis() ) other += ", ...";
  other += ")" + leftover;
  if ( signature()->isConst() ) other += " const";
  if ( linkage() == LK_C ) other += " C";

  // Implementation B:
  // use signature
  const string strsig = signature()->name();
  int idx = strsig.size() - 1;
  int len = idx;
  if ( signature()->returnType()->kind() == TY_PTR &&
       signature()->returnType()->pointedToType()->kind() == TY_FUNC ) {
    // return type is function pointer
    const char *sig = strsig.c_str();
    const char *ptr = strstr(sig, "(*");
    idx = (ptr - sig) + 2;
  } else {
    // normal function
    int paren = 0;
    while ( idx >= 0 ) {
      if ( strsig[idx] == ')' ) {
        ++paren;
      } else if ( strsig[idx] == '(' ) {
        --paren;
        if ( paren == 0 ) break;
      }
      --idx;
    }
  }
  if ( idx >= 0 ) {
    if ( linkage() == LK_C && strsig[len-1]!=' ' && strsig[len]!='C' ) {
      // in C mode, edg doesn't add linkage specifiers
      fullName(strsig.substr(0,idx) + fullName()
                                    + strsig.substr(idx, len-idx+1) + " C");
    } else
      fullName(strsig.substr(0,idx) + fullName()
                                    + strsig.substr(idx, len-idx+1));
  }

  // Compare A vs. B
  if ( other != fullName() ) {
    cerr << "other *** '" << other << "'" << endl;
    cerr << "full  *** '" << fullName() << "'" << endl;
  }
}

ostream& pdbRoutine::print(ostream& ostr) const {
  func_t k;
 
  pdbItem::print(ostr);
  ostr << "rsig ty#" << signature()->id() << "\n";
  ostr << "rlink " << toName(linkage()) << "\n";
  ostr << "rstore " << toName(storageClass()) << "\n";
  if ( (k = specialKind()) != FU_NA )
    ostr << "rkind " << toName(k) << "\n";
  ostr << "rvirt " << toName(virtuality()) << "\n";
  if ( covariantReturnVirtualOverride() ) ostr << "rcrvo T\n";
  if ( isInline() ) ostr << "rinline T\n";
  if ( isCompilerGenerated() ) ostr << "rcgen T\n";
  if ( isExplicitCtor() ) ostr << "rexpl T\n";
  if ( isStatic() ) ostr << "rstatic T\n";
  pdbTemplateItem::print(ostr);
  const pdbRoutine::callvec c = callees();
  for (pdbRoutine::callvec::const_iterator ct=c.begin(); ct!=c.end(); ++ct)
    ostr << "rcall " << **ct << "\n";
  pdbFatItem::print(ostr);
  return ostr << endl;
}

void pdbRoutine::adjustPtrs(PDB* p) {
  const pdbRoutine* r;
  const pdbFile* fi;

  pdbTemplateItem::adjustPtrs(p);
  if ( sig->newId() > UNIQUE ) sig = p->getTypeMap()[sig->newId()];
  for( callvec::iterator it = c.begin(); it!=c.end(); ++it) {
    r = (*it)->call();
    if ( r->newId() > UNIQUE ) (*it)->call(p->getRoutineMap()[r->newId()]);
    if ( (fi=(*it)->file()) && (fi->newId() > UNIQUE) )
      (*it)->file(p->getFileMap()[fi->newId()]);
  }
}

pdbSimpleItem::dupl_t pdbRoutine::findDuplicate(pdbSimpleItem* r) {
  pdbRoutine* rhs = dynamic_cast<pdbRoutine*> (r);
  if ( fullName() == rhs->fullName() ) {
    if ( storageClass() == ST_EXT && rhs->storageClass() != ST_EXT )
      return OLDDUPL;
    else
      return NEWDUPL;
  } else {
      return NODUPL;
  }
}
