/*************************************************************************/
/* DUCTAPE Version 1.0                                                   */
/* Copyright (C) 1998                                                    */
/* Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik */
/* Los Alamos National Laboratory, Advanced Computing Laboratory         */
/*************************************************************************/

#include "pdbType.h"
#include "pdbClass.h"

const pdbClass* pdbType::NULLEXCEP = (const pdbClass*)1;

void pdbType::process(PDB *p) {
  // ??? name is fullName, because of bug in frontend
  // ??? to fix this bug, try to correct name
  // but only for enums and typedefs
  size_t pos;
  if ( kind()==TY_ENUM ||
      (kind()==TY_TREF && qualifiers().size()==0
                       && name()[name().size()-1]!='>') ) {
    if ( (pos = name().rfind("::")) != STR_NPOS ) {
      name(name().substr(pos+2, STR_NPOS));
    }
  }
  pdbItem::process(p);

  // is Const?
  const qualvec& q = qualifiers();
  if ( q.size() ) {
    for (qualvec::const_iterator it=q.begin(); it!=q.end(); ++it) {
      if ( (*it) == QL_CONST ) {
        isConst(true);
        break;
      }
    }
  }
}

static void printType(ostream& ostr, char *attr, const pdbType* ty,
                      const char *extra = "") {
  const pdbClass *cl;

  if ( (cl = ty->isClass()) != 0 )
    ostr << attr << "cl#" << cl->id() << extra << "\n";
  else
    ostr << attr << "ty#" << ty->id() << extra << "\n";
}

ostream& pdbType::print(ostream& ostr) const {
  qualvec q;

  pdbItem::print(ostr);
  ostr << "ykind " << toName(kind()) << "\n";
  switch ( kind() ) {
  case TY_INT:
  case TY_BOOL:
  case TY_WCHAR:
    ostr << "yikind " << toName(integerType()) << "\n";
    if ( isSigned() ) ostr << "ysigned T" << "\n";
    break;
  case TY_ENUM: {
    ostr << "yikind " << toName(integerType()) << "\n";
    for (enumvec::const_iterator eit=en.begin(); eit!=en.end(); ++eit)
      ostr << "yenum " << (*eit).id() << " " << (*eit).val() << "\n";
    }
    break;
  case TY_FLOAT:
    ostr << "yfkind " << toName(floatType()) << "\n";
    break;
  case TY_PTR:
    printType(ostr, "yptr ", pointedToType());
    break;
  case TY_REF:
    printType(ostr, "yref ", referencedType());
    break;
  case TY_ARRAY:
    printType(ostr, "yelem ", elementType());
    ostr << "ynelem " << numElements() << "\n";
    break;
  case TY_TREF:
    printType(ostr, "ytref ", baseType());
    q = qualifiers();
    if ( q.size() ) {
      ostr << "yqual";
      for (qualvec::iterator it=q.begin(); it!=q.end(); ++it)
        ostr << " " << toName(*it);
      ostr << "\n";
    }
    break;
  case TY_FUNC: {
    printType(ostr, "yrett ", returnType());
    for (argvec::const_iterator ait=a.begin(); ait!=a.end(); ++ait) {
      printType(ostr, "yargt ", ait->type(), ait->hasDefault() ? " T" : " F");
    }
    if ( hasEllipsis() ) ostr << "yellip T" << "\n";
    if ( isConst() ) ostr << "yqual const\n";
    for (classvec::const_iterator xit=es.begin(); xit!=es.end(); ++xit)
      if ( (*xit) == NULLEXCEP )
        ostr << "yexcep NULL\n";
      else
        ostr << "yexcep cl#" << (*xit)->id() << "\n";
    }
    break;
  case TY_PTRMEM:
    ostr << "ympclass cl#" << memberPointerClass()->id() << "\n";
    printType(ostr, "ymptype ", memberPointerType());
    break;
  default:
    // nothing
    break;
  }
  return ostr << endl;
}

void pdbType::adjustPtrs(PDB* pdb) {
  pdbItem::adjustPtrs(pdb);
  if ( e && (e->newId() > UNIQUE) ) e = pdb->getTypeMap()[e->newId()];
  if ( p && (p->newId() > UNIQUE) ) p = pdb->getTypeMap()[p->newId()];
  if ( b && (b->newId() > UNIQUE) ) b = pdb->getTypeMap()[b->newId()];
  if ( r && (r->newId() > UNIQUE) ) r = pdb->getTypeMap()[r->newId()];
  if ( mt && (mt->newId() > UNIQUE) ) mt = pdb->getTypeMap()[mt->newId()];
  if ( mc && (mc->newId() > UNIQUE) ) mc = pdb->getClassMap()[mc->newId()];
  if ( c && (c->newId() > UNIQUE) ) c = pdb->getClassMap()[c->newId()];
  for (argvec::iterator ait=a.begin(); ait!=a.end(); ++ait) {
    if ( ait->type()->newId() > UNIQUE )
      ait->type(pdb->getTypeMap()[ait->type()->newId()]);
  }
  for (classvec::iterator xit=es.begin(); xit!=es.end(); ++xit) {
    if ( ((*xit) != NULLEXCEP) && ((*xit)->newId() > UNIQUE) )
      *xit = pdb->getClassMap()[(*xit)->newId()];
  }
}

pdbSimpleItem::dupl_t pdbType::findDuplicate(pdbSimpleItem* t) {
  pdbType* rhs = dynamic_cast<pdbType*> (t);
  if ( isClass() && rhs->isClass() ) 
    // if both are class types, compare class names
    return ( isClass()->fullName() == rhs->isClass()->fullName() ?
             NEWDUPL : NODUPL );
  else if ( isClass() || rhs->isClass() )
    // if only one of them is class type, they are different
    return NODUPL;
  else
    // otherwise compare full names as usual
    return ( fullName() == rhs->fullName() ? NEWDUPL : NODUPL );
}
