/*************************************************************************/
/* DUCTAPE Version 1.0                                                   */
/* Copyright (C) 1998                                                    */
/* Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik */
/* Los Alamos National Laboratory, Advanced Computing Laboratory         */
/*************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#ifdef _OLD_HEADER_
# include <fstream.h>
# include <set.h>
# include <algo.h>
#else
# include <fstream>
  using std::ifstream;
  using std::ofstream;
# include <set>
  using std::set;
# include <algorithm>
  using std::sort;
#endif
#include "pdbAll.h"

static const char *keywords[73] = {
  "and", "and_eq", "asm", "auto", "bitand", "bitor", "bool", "break",
  "case", "catch", "char", "class", "compl", "const", "const_cast",
  "continue", "default", "delete", "do", "double", "dynamic_cast", "else",
  "enum", "explicit", "extern", "false", "float", "for", "friend", "goto",
  "if", "inline", "int", "long", "mutable", "namespace", "new", "not",
  "not_eq", "operator", "or", "or_eq", "private", "protected", "public",
  "register", "reinterpret_cast", "return", "short", "signed", "sizeof",
  "static", "static_cast", "struct", "switch", "template", "this", "throw",
  "true", "try", "typedef", "typeid", "typename", "union", "unsigned",
  "using", "virtual", "void", "volatile", "wchar_t", "while", "xor", "xor_eq"
};

struct ltstr {
  bool operator()(const char* s1, const char* s2) const {
    return strcmp(s1, s2) < 0;
  }
};

static bool itemCmp(const pdbSimpleItem* i1, const pdbSimpleItem* i2) {
  return i1->name() < i2->name();
}

static void out(ostream& ostr, char ch) {
  if ( ch == '<' )
    ostr << "&lt;";
  else if ( ch == '>' )
    ostr << "&gt;";
  else if ( ch == '&' )
    ostr << "&amp;";
  else
    ostr << ch;
}

static void outStr(ostream& ostr, const string& str) {
  for (unsigned i=0; i<str.size(); ++i) out(ostr, str[i]);
}

static bool isSysFile(const char *fname) {
  return (strncmp(fname, "/usr", 4)==0) || (strncmp(fname, "/opt", 4)==0);
}

static void generateIndex(PDB& pdb, ostream& ostr, const char *pdbfile) {
  cout << "Generating index.html ..." << endl;
  ostr << "<HTML>\n<HEAD>\n<TITLE>" << pdbfile << " Index</TITLE>\n</HEAD>\n"
       << "<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\""
       << "LINK=\"#000000\" ALINK=\"#000000\" VLINK=\"#000000\">\n"
       << "<H1 ALIGN=\"CENTER\">Index of " << pdbfile << "</H1>\n"
       << "<HR>\n";

  ostr << "<H2>Files:</H2>\n<UL>\n";
  PDB::filevec files = pdb.getFileVec();
  sort(files.begin(), files.end(), itemCmp);
  for (PDB::filevec::const_iterator it=files.begin(); it!=files.end(); ++it) {
    if ( isSysFile((*it)->name().c_str()) ) continue;
    const char *outfile = ((*it)->name() + ".html").c_str();
    ostr << "<LI><A HREF=\"" << outfile << "\">"
         << (*it)->name() << "</A>\n";
  }
  ostr << "</UL>\n";

  ostr << "<H2>Classes and Methods:</H2>\n<UL>\n";
  PDB::classvec classes = pdb.getClassVec();
  sort(classes.begin(), classes.end(), itemCmp);
  for (PDB::classvec::const_iterator cit=classes.begin();
       cit!=classes.end(); ++cit) {
    if ( isSysFile((*cit)->location().file()->name().c_str()) ) continue;
    ostr << "<LI><A HREF=\"" << (*cit)->location().file()->name()
         << ".html#cl" << (*cit)->id() << "\">";
    outStr(ostr, (*cit)->fullName());
    ostr << "</A>\n";
    pdbClass::methodvec methods = (*cit)->funcMembers();
    if ( methods.size() ) {
      ostr << "<UL>\n";
      for (pdbClass::methodvec::iterator mit = methods.begin();
           mit!=methods.end(); ++mit) {
        if ( (*mit)->func()->isCompilerGenerated() ) {
          ostr << "<LI>";
          outStr(ostr, (*mit)->func()->fullName());
          ostr << "\n";
        } else {
          ostr << "<LI><A HREF=\"" << (*mit)->func()->location().file()->name()
               << ".html#ro" << (*mit)->func()->id() << "\">";
          outStr(ostr, (*mit)->func()->fullName());
          ostr << "</A>\n";
        }
      }
      ostr << "</UL>\n";
    }
  }
  ostr << "</UL>\n";

  ostr << "<H2>Templates:</H2>\n<UL>\n";
  PDB::templatevec templates = pdb.getTemplateVec();
  sort(templates.begin(), templates.end(), itemCmp);
  for (PDB::templatevec::const_iterator tit=templates.begin();
       tit!=templates.end(); ++tit) {
    if ( isSysFile((*tit)->location().file()->name().c_str()) ) continue;
    ostr << "<LI><A HREF=\"" << (*tit)->location().file()->name()
         << ".html#te" << (*tit)->id() << "\">";
    outStr(ostr, (*tit)->fullName());
    ostr << "</A>\n";
  }
  ostr << "</UL>\n";

  ostr << "<H2>Functions:</H2>\n<UL>\n";
  PDB::routinevec routines = pdb.getRoutineVec();
  sort(routines.begin(), routines.end(), itemCmp);
  for (PDB::routinevec::const_iterator rit=routines.begin();
       rit!=routines.end(); ++rit) {
    if ( ! (*rit)->parentClass() ) {
      if ( (*rit)->isCompilerGenerated() ) {
        ostr << "<LI>";
        outStr(ostr, (*rit)->fullName());
        ostr << "\n";
      } else {
        if ( isSysFile((*rit)->location().file()->name().c_str()) ) continue;
        ostr << "<LI><A HREF=\"" << (*rit)->location().file()->name()
             << ".html#ro" << (*rit)->id() << "\">";
        outStr(ostr, (*rit)->fullName());
        ostr << "</A>\n";
      }
    }
  }
  ostr << "</UL>\n";

  ostr << "<H2>Namespaces:</H2>\n<UL>\n";
  PDB::namespacevec namespaces = pdb.getNamespaceVec();
  sort(namespaces.begin(), namespaces.end(), itemCmp);
  for (PDB::namespacevec::const_iterator nit=namespaces.begin();
       nit!=namespaces.end(); ++nit) {
    if ( !(*nit)->location().file() ||
         isSysFile((*nit)->location().file()->name().c_str()) ) continue;
    ostr << "<LI><A HREF=\"" << (*nit)->location().file()->name()
         << ".html#na" << (*nit)->id() << "\">";
    outStr(ostr, (*nit)->fullName());
    ostr << "</A>\n";
  }
  ostr << "</UL>\n";

  ostr << "<HR>\n</BODY>\n</HTML>\n";
}

struct itemRef {
  itemRef(const pdbItem *i, bool isT) : item(i), isTarget(isT) {
    line = i->location().line();
    col  = i->location().col();
  }
  itemRef(const pdbItem *i, bool isT, int l, int c)
         : item(i), isTarget(isT), line(l), col(c) {}
  const pdbItem *item;
  bool     isTarget;
  int      line;
  int      col;
};

static bool locCmp(const itemRef* r1, const itemRef* r2) {
  if ( r1->line == r2->line )
    return r1->col < r2->col;
  else
    return r1->line < r2->line;
}

void getReferences(vector<itemRef *>& itemvec, PDB& pdb, pdbFile *file) {
  for (PDB::classvec::const_iterator it=pdb.getClassVec().begin();
    it!=pdb.getClassVec().end(); ++it) {
    if ( (*it)->location().file() == file ) {
      itemvec.push_back(new itemRef(*it, true));
      for (pdbClass::fieldvec::const_iterator dit=(*it)->dataMembers().begin();
           dit!=(*it)->dataMembers().end(); ++dit) {
        if ( (*dit)->location().file() && (*dit)->kind()!=pdbItem::M_TYPE
                                       && !((*dit)->type()->isClass()) )
          itemvec.push_back(new itemRef(*dit, true));
      }

    }
  }
  PDB::routinevec routines = pdb.getRoutineVec();
  for (PDB::routinevec::const_iterator rit=routines.begin();
       rit!=routines.end(); ++rit) {
    if ( (*rit)->location().file() == file && !(*rit)->isCompilerGenerated()) {
      itemvec.push_back(new itemRef(*rit, true));
      for (pdbRoutine::callvec::iterator cit=(*rit)->callees().begin();
           cit!=(*rit)->callees().end(); ++cit) {
        if ( (*cit)->line() > 0 )
          itemvec.push_back(
             new itemRef((*cit)->call(), false, (*cit)->line(), (*cit)->col()));
      }
    }
  }
  PDB::templatevec templates = pdb.getTemplateVec();
  for (PDB::templatevec::const_iterator tit=templates.begin();
       tit!=templates.end(); ++tit) {
    if ( (*tit)->location().file() == file ) {
      itemvec.push_back(new itemRef(*tit, true));
    }
  }
  PDB::typevec types = pdb.getTypeVec();
  for (PDB::typevec::const_iterator yit=types.begin();
       yit!=types.end(); ++yit) {
    if ( (*yit)->location().file() == file &&
        (((*yit)->kind()==pdbItem::TY_TREF && (*yit)->qualifiers().size()==0)
       || (*yit)->kind()==pdbItem::TY_ENUM) ) {
      itemvec.push_back(new itemRef(*yit, true));
    }
  }
  PDB::namespacevec namespaces = pdb.getNamespaceVec();
  for (PDB::namespacevec::const_iterator nit=namespaces.begin();
       nit!=namespaces.end(); ++nit) {
    if ( (*nit)->location().file() == file ) {
      itemvec.push_back(new itemRef(*nit, true));
    }
  }
  sort(itemvec.begin(), itemvec.end(), locCmp);
}

static const pdbItem* setNextItem(vector<itemRef *>& itemvec,
                           int& curr, int& nextLine, int& nextCol) {
  ++curr;
  if ( curr >= itemvec.size() ) {
    nextLine = -1;
    nextCol  = -1;
    return 0;
  } else {
    nextLine = itemvec[curr]->line;
    nextCol  = itemvec[curr]->col;
    return itemvec[curr]->item;
  }
}

static string getQualifiedName(const pdbItem* item) {
  string fn = item->name();
  const pdbClass *cptr = item->parentClass();
  while ( cptr ) {
    fn = cptr->name() + "::" + fn;
    cptr = cptr->parentClass();
  }
  const pdbNamespace *nptr = item->parentNSpace();
  while ( nptr ) {
    fn = nptr->name() + "::" + fn;
    nptr = nptr->parentNSpace();
  }
  return fn;
}

static int nextMultiple = -1;

const int INBUF_SIZE = 2048;

static bool processFile(PDB& pdb, pdbFile* f,
                      const set<const char *, ltstr>& isKeyword, 
                      const map<const char *, const pdbItem *, ltstr>& isItem) {
  string file(f->name());
  static char inbuf[INBUF_SIZE];
  int line = 1;
  int pos = 0;
  bool inComment = false;
  bool inPreStmt = false;

  // do not try to process system files
  if ( isSysFile(f->name().c_str()) ) return true;

  // open source file
  ifstream istr(file.c_str());
  if (!istr) {
    cerr << "Error: Cannot open '" << file << "'" << endl;
    return false;
  }
  cout << "Processing " << file << " ..." << endl;

  // open outfile for HTML version of source file
  const char *outfile = (file + ".html").c_str();
  ofstream ostr(outfile);
  if (!ostr) {
    cerr << "Error: Cannot open '" << outfile << "'" << endl;
    return false;
  }

  // determine pathname prefix
  unsigned spos = STR_NPOS;
  string fprefix("");
  if ( (spos = file.rfind('/')) != STR_NPOS ) fprefix = file.substr(0,spos+1);

  // initialize reference vector
  vector<itemRef *> itemvec;
  getReferences(itemvec, pdb, f);
  int curr = -1;
  int nextLine = -1;
  int nextCol  = -1;
  const pdbItem* nextItem = 0;
  nextItem = setNextItem(itemvec, curr, nextLine, nextCol);

  // write header
  ostr << "<HTML>\n<HEAD>\n<TITLE>" << file << "</TITLE>\n</HEAD>\n"
       << "<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\""
       << "LINK=\"#000000\" ALINK=\"#000000\" VLINK=\"#000000\">\n"
       << "<H1>File: " << file << "</H1>\n<HR>\n<PRE>\n";

  // parse file
  while ( istr.getline(inbuf, INBUF_SIZE) ) {
    if ( !inComment && inbuf[0] == '#' ) {    /*??? ERROR # first non-white*/
      // preprocessor statement
      bool isInc = false;
      bool isIncRef = false;
      ostr << "<FONT COLOR=#004000>";
      while ( inbuf[pos] ) {
        if ( inbuf[pos] == 'i' && strncmp(inbuf+pos, "include", 7)==0 ) {
          isInc = true;
          ostr << 'i';
        } else if ( isIncRef && (inbuf[pos]=='\"' || inbuf[pos]=='>') ) {
          ostr << "</A>" << inbuf[pos];
        } else if ( isInc && (inbuf[pos]=='\"' || inbuf[pos]=='<') ) {
          out(ostr, inbuf[pos]);
          int len = 0;
          while ( inbuf[pos+len+1]!='\"' && inbuf[pos+len+1]!='>' ) ++len;
          for (PDB::filevec::const_iterator it=pdb.getFileVec().begin();
               it!=pdb.getFileVec().end(); ++it) {
            const char *fname = ((*it)->name()).c_str();
            int flen = strlen(fname);
            if ( len<=flen  &&  strncmp(inbuf+pos+1, fname+flen-len, len)==0 ) {
              ostr << "<A HREF=\"" << fname << ".html\">";
              isIncRef = true;
              break;
            }
          }
        } else {
          out(ostr, inbuf[pos]);
        }
        ++pos;
      }
      if ( inbuf[pos-1] == '\\' )
        inPreStmt = true;
      else
        ostr << "</FONT>";
    } else if ( inPreStmt ) {
      // preprocessor statement continuation
      while ( inbuf[pos] ) out(ostr, inbuf[pos++]);  
      if ( inbuf[pos-1] == '\\' )
        inPreStmt = true;
      else {
        ostr << "</FONT>";
        inPreStmt = false;
      }
    } else {
      while ( inbuf[pos] ) {
        if ( inComment ) {
          /* look for comment end */
          if ( inbuf[pos] == '*' && inbuf[pos+1] == '/' ) {
            ostr << "*/</FONT>";
            inComment = false;
            pos += 2;
          } else {
            out(ostr, inbuf[pos++]);
          }
        } else if ( inbuf[pos] == '/' ) {
          pos++;
          if ( inbuf[pos] == '/' ) {
            // comments
            pos++;
            ostr << "<FONT COLOR=#0000c0>//";
            while ( inbuf[pos] ) { out(ostr, inbuf[pos++]); }
            ostr << "</FONT>";
          } else if ( inbuf[pos] == '*' ) {
            /* comment start */
            pos++;
            ostr << "<FONT COLOR=#0000c0>/*";
            inComment = true;
          } else {
            ostr << '/';
          }
        } else if ( inbuf[pos] == '\"' ) {
          // character string constant
          ostr << "<FONT COLOR=#ff3f00>\"";
          do {
            pos++;
            if ( inbuf[pos] == '\\' ) {
              ostr << '\\';
              pos++;
              if ( inbuf[pos] == '\"' ) {
                ostr << '\"';
                pos++;
              }
            }
            out(ostr, inbuf[pos]);
          }
          while ( inbuf[pos] != '\"' );
          ostr << "</FONT>";
          pos++;
        } else if ( inbuf[pos] == '\'' ) {
          // character constant
          ostr << "<FONT COLOR=#ff3f00>\'";
          do {
            pos++;
            if ( inbuf[pos] == '\\' ) {
              ostr << '\\';
              pos++;
              if ( inbuf[pos] == '\'' ) {
                ostr << '\'';
                pos++;
              }
            }
            out(ostr, inbuf[pos]);
          }
          while ( inbuf[pos] != '\'' );
          ostr << "</FONT>";
          pos++;
        } else {
          if ( nextLine == line && nextCol == (pos+1) ) {
            // hyperlink item
            // ignore multiple entries of the same thing
            while ( curr < (itemvec.size()-1) &&
                    nextItem == itemvec[curr+1]->item &&
                    nextLine == itemvec[curr+1]->line &&
                    nextCol  == itemvec[curr+1]->col ) {
              nextItem = setNextItem(itemvec, curr, nextLine, nextCol);
            }

            // multiple target or link?
            bool isMult = false;
            if ( curr < (itemvec.size()-1) &&
                 nextItem != itemvec[curr+1]->item &&
                 nextLine == itemvec[curr+1]->line &&
                 nextCol  == itemvec[curr+1]->col ) {
              isMult = true;
            }

            if ( itemvec[curr]->isTarget ) {
/*???*/       const pdbTemplate *t = dynamic_cast<const pdbTemplate*>(nextItem);
              int numAnchor = 0;
              const string& itemName = nextItem->name();

              ostr << "<FONT COLOR=#c00000>";
              do {
                ostr << "<A NAME=\""
                     << nextItem->desc()[0] << nextItem->desc()[1]
                     << nextItem->id() << "\">";
                ++numAnchor;
                nextItem = setNextItem(itemvec, curr, nextLine, nextCol);
              } while ( isMult &&
                        nextItem != itemvec[curr-1]->item &&
                        nextLine == itemvec[curr-1]->line &&
                        nextCol  == itemvec[curr-1]->col );
              if (t)
                for (int i=0; i<8; ++i)
                  out(ostr, inbuf[pos++]);
              else
                for (unsigned i=0; i<itemName.size(); ++i) {
                  if ( itemName[i] == '<' ) break;
                  out(ostr, inbuf[pos++]);
                }
              while ( numAnchor-- ) ostr << "</A>";
              ostr << "</FONT>";
            } else {
              int len;
              char multFile[1024];

              if ( isMult ) {
                // multiple targets for the same anchor
                // -> generate auxillary target table file
                sprintf(multFile, "%s__mult%d.html",
                        fprefix.c_str(), ++nextMultiple);

                // generate link to auxillary target table file
                ostr << "<A HREF=\"" << multFile << "\">";

              } else {
                ostr << "<A HREF=\"" << nextItem->location().file()->name()
                     << ".html#" << nextItem->desc()[0] << nextItem->desc()[1]
                     << nextItem->id() << "\">";
              }
              if ( strncmp(nextItem->name().c_str(), "operator", 8) == 0 ) {
                const char *opname = nextItem->name().c_str() + 8;
                while ( *opname == ' ' ) ++opname;
                if ( *opname == '[' || *opname == '(' )
                  len = 1;
                else
                  len = strlen(opname);
              } else {
                len = nextItem->name().size();
                if ( strncmp(inbuf+pos, nextItem->name().c_str(), len) != 0) {
                  // name doesn't match: try qualified name
                  const char *qname = getQualifiedName(nextItem).c_str();
                  if ( strncmp(inbuf+pos, qname, strlen(qname)) == 0)
                    len = strlen(qname);
                  else {
                    /*??? qualified name with template class*/
                    //cerr << "WARNING: non-matching name " << nextItem->name()
                    //     << " at " << nextLine << " " << nextCol << endl;
                  }
                }
              }
              for (int i=0; i<len; ++i)
                out(ostr, inbuf[pos++]);
              ostr << "</A>";

              if ( isMult ) {
                // create auxillary target table file
                ofstream mul(multFile);
                if (!mul) {
                  cerr << "Error: Cannot open '" << multFile << "'" << endl;
                  return false;
                }
                mul << "<HTML>\n<HEAD>\n<TITLE>"
                    << multFile << "</TITLE>\n</HEAD>\n"
                    << "<BODY BGCOLOR=\"#ffffff\" TEXT=\"#000000\""
                    << "LINK=\"#000000\" ALINK=\"#000000\" VLINK=\""
                    << "#000000\">\n";
                mul << "<H1>Target Table for " << nextItem->name() << "</H1>\n";

                // add targets to auxillary target table file
                mul << "<UL>\n";
                do {
                  mul << "<LI><A HREF=\"" << nextItem->location().file()->name()
                      << ".html#" << nextItem->desc()[0] << nextItem->desc()[1]
                      << nextItem->id() << "\">";
                  outStr(mul, nextItem->fullName());
                  mul << "</A>\n";
                  nextItem = setNextItem(itemvec, curr, nextLine, nextCol);
                } while ( nextItem != itemvec[curr-1]->item &&
                          nextLine == itemvec[curr-1]->line &&
                          nextCol  == itemvec[curr-1]->col );
                mul << "</UL>\n";
                mul << "</BODY>\n</HTML>\n";
              } else {
                nextItem = setNextItem(itemvec, curr, nextLine, nextCol);
              }
            }

          } else if ( isalpha(inbuf[pos]) ) {
            /*???*/
            if ( nextItem && ((nextLine == line && nextCol < (pos+1)) ||
                                  (nextLine < line)) ) {
            cerr << "ERROR: skipping " << nextItem->desc() << "#"
                 << nextItem->id() << " " << nextItem->name() << " at "
                 << nextLine << ":" <<nextCol << " -> " << line << ":"
                 << (pos+1) << endl;
              nextItem = setNextItem(itemvec, curr, nextLine, nextCol);
            }

            // identifier  /*??? _identifier?*/
            map<const char *, const pdbItem *, ltstr>::const_iterator it;
            const pdbItem *item;

            int startPos = pos;
            while ( isalnum(inbuf[pos]) || inbuf[pos]=='_' ) pos++;
            char saveCh = inbuf[pos];
            inbuf[pos] = '\0';
            if ( isKeyword.find(inbuf+startPos) != isKeyword.end() ) {
              // keyword
              ostr << "<STRONG>" << inbuf+startPos << "</STRONG>";
            } else if ( (it = isItem.find(inbuf+startPos)) != isItem.end() &&
                        (item = (*it).second) != 0) {
              // class, enum, or enum value name
              ostr << "<A HREF=\"" << item->location().file()->name()
                   << ".html#" << item->desc()[0] << item->desc()[1]
                   << item->id() << "\">";
              outStr(ostr, item->name());
              ostr << "</A>";
            } else {
              ostr << inbuf+startPos;
            }
            inbuf[pos] = saveCh;
          } else if ( isdigit(inbuf[pos]) ) {
            // numerical constant?
            char *eptr1, *eptr2;
            strtod(inbuf+pos, &eptr1);     // decimal, octal, float?
            strtol(inbuf+pos, &eptr2, 16); // hexadecimal?
            if ( eptr1 != (inbuf+pos) || eptr2 != (inbuf+pos) ) {
              int len = (eptr1 < eptr2) ? eptr2-(inbuf+pos) : eptr1-(inbuf+pos);
              ostr << "<FONT COLOR=#ff3f00>";
              for (int i=0; i<len; ++i) out(ostr, inbuf[pos++]);
              ostr << "</FONT>";
            } else {
              out(ostr, inbuf[pos++]);
            }
          } else {
            /*???*/
            if ( nextItem && ((nextLine == line && nextCol < (pos+1)) ||
                                  (nextLine < line)) ) {
            cerr << "ERROR: skipping " << nextItem->desc() << "#"
                 << nextItem->id() << " " << nextItem->name() << " at "
                 << nextLine << ":" <<nextCol << " -> " << line << ":"
                 << (pos+1) << endl;
              nextItem = setNextItem(itemvec, curr, nextLine, nextCol);
            }
            out(ostr, inbuf[pos++]);
          }
        }
      }
    }
    ostr << endl;
    ++line;
    pos = 0;
  }

  // write footer
  ostr << "</PRE>\n";
  //ostr << "<HR>\n";
  //ostr << "Back to <A HREF=\"index.html\">index</A>\n";
  ostr << "</BODY>\n</HTML>\n";

  for (curr=0; curr<itemvec.size(); ++curr) delete itemvec[curr];
  return true;
}

int main(int argc, char *argv[]) {
  if ( argc != 2 ) {
    cerr << "usage: " << argv[0] << " pdbfile" << endl;
    return 1;
  }

  // open and initialize program database
  PDB pdb(argv[1]);
  if ( !pdb ) return 1;

  // initialize keyword and item lookup tables
  set<const char *, ltstr> isKeyword(keywords, keywords+73);
  map<const char *, const pdbItem *, ltstr> isItem;
  map<const char *, const pdbItem *, ltstr>::iterator iit;
  PDB::classvec& c = pdb.getClassVec();
  for (PDB::classvec::const_iterator cit=c.begin(); cit!=c.end(); ++cit) {
    if ( !(*cit)->isTemplate() ) {
      if ( (iit = isItem.find((*cit)->name().c_str())) != isItem.end() ) {
        //cerr << "WARNING: duplicate item " << (*cit)->name() << endl;
        (*iit).second = 0;
      } else {
        isItem[(*cit)->name().c_str()] = (*cit);
      }
    }
  }
  PDB::typevec& t = pdb.getTypeVec();
  for (PDB::typevec::const_iterator tit=t.begin(); tit!=t.end(); ++tit) {
    if ( ((*tit)->kind()==pdbItem::TY_TREF && (*tit)->qualifiers().size()==0)
       || (*tit)->kind()==pdbItem::TY_ENUM ) {
      if ( (iit = isItem.find((*tit)->name().c_str())) != isItem.end() ) {
        //cerr << "WARNING: duplicate item " << (*tit)->name() << endl;
        (*iit).second = 0;
      } else {
        isItem[(*tit)->name().c_str()] = (*tit);
      }
    }
  }
  PDB::namespacevec& n = pdb.getNamespaceVec();
  for (PDB::namespacevec::const_iterator nit=n.begin(); nit!=n.end(); ++nit) {
    if ( (*nit)->location().file() ) {
      if ( (iit = isItem.find((*nit)->name().c_str())) != isItem.end() ) {
        //cerr << "WARNING: duplicate item " << (*tit)->name() << endl;
        (*iit).second = 0;
      } else {
        isItem[(*nit)->name().c_str()] = (*nit);
      }
    }
  }

  // generate index files
  {
    ofstream ostr("index.html");
    if (!ostr) {
      cerr << "Error: Cannot open 'index.html'" << endl;
      return 1;
    }
    generateIndex(pdb, ostr, argv[1]);
  }

  // generate HTML version of every source file
  for (PDB::filevec::const_iterator it=pdb.getFileVec().begin();
       it!=pdb.getFileVec().end(); ++it) {
    if ( ! processFile(pdb, *it, isKeyword, isItem) ) return 1;
  }
  return 0;
}
